//
//  pgj.java
//  pgj
//
//  Created by Brian Olson on Mon Apr 26 2004.
//  Copyright (c) 2004 Brian Olson. All rights reserved.
//
import java.util.*;
import java.net.*;
import java.io.*;

public class pgj implements Runnable {

	static final String defaultServerName = "localhost";
	static final int defaultServerPort = 5432;

	Socket s;
	//InputStream is;
	DataInputStream is;
	OutputStream os;

	/**
		@param server something like "localhost" or "db.a.com"
		@param port 5432 is common
		@param params "name","value","name","value",...
		*/
	public void openConnection( String server, int port, String[] params ) throws Exception {
		s = new Socket( server, port );
		os = s.getOutputStream();
		//is = s.getInputStream();
		is = new DataInputStream( s.getInputStream() );
		byte[][] msgParts = new byte[params.length][];
		int size = 9; // size + version + zero-byte-at-end

		(new Thread(this)).start();
		for ( int i = 0; i < params.length; i++ ) {
			byte[] tb;
			tb = params[i].getBytes("UTF8");
			msgParts[i] = tb;
			size += tb.length + 1;
		}
		byte[] msg = new byte[size];
		if ( true ) {
			intToByteArray( msg, 0, size );
		} else {
		msg[0] = (byte)((size >> 24) & 0xff);
		msg[1] = (byte)((size >> 16) & 0xff);
		msg[2] = (byte)((size >>  8) & 0xff);
		msg[3] = (byte)((size      ) & 0xff);
		}
		// version
		msg[4] = 0;
		msg[5] = 3;
		msg[6] = 0;
		msg[7] = 0;

		int msgPos = 8;

		for ( int i = 0; i < msgParts.length; i++ ) {
			System.arraycopy( msgParts[i], 0, msg, msgPos, msgParts[i].length );
			msgPos += msgParts[i].length;
			msg[msgPos] = 0;
			msgPos++;
		}
		msg[msgPos] = 0;
		os.write( msg );
	}
	static final byte[] terminateMessage = { (byte)'X', 0, 0, 0, 4 };
	static final byte[] syncMessage = { (byte)'S', 0, 0, 0, 4 };
	/** send the terminate message and close the secket */
	public void closeConnection() throws Exception {
		if ( s == null ) return;
		os.write( terminateMessage );
		os.flush();
		is.close();
		os.close();
		s.close();
	}
	/** 
		@param statementName null for unnamed statement, otherwise a name for preparing a statement 
		@param query sql query text
		@param paramTypes oid type of parameters */
	public void parse( String statementName, String query, int[] paramTypes ) throws Exception {
		int len = 6; // size + numParams
		byte[] name;
		byte[] q;
		if ( statementName == null ) {
			name = null;
			len += 1;
		} else {
			name = statementName.getBytes("UTF8");
			len += name.length + 1;
		}
		q = query.getBytes("UTF8");
		len += q.length + 1;
		if ( paramTypes != null ) {
			len += paramTypes.length * 4;
		}
		byte[] buf = new byte[len+1];
		int bp = 0;
		buf[bp] = (byte)'P';
		bp++;
		intToByteArray( buf, bp, len );
		bp += 4;
		if ( name != null ) {
			System.arraycopy( name, 0, buf, bp, name.length );
			bp += name.length;
		}
		buf[bp] = 0; bp++;
		System.arraycopy( q, 0, buf, bp, q.length );
		bp += q.length;
		buf[bp] = 0; bp++;
		if ( paramTypes != null ) {
			shortToByteArray( buf, bp, (short)paramTypes.length );
			bp += 2;
			for ( int i = 0; i < paramTypes.length; i++ ) {
				intToByteArray( buf, bp, paramTypes[i] );
				bp += 4;
			}
		} else {
			shortToByteArray( buf, bp, 0 );
			bp += 2;
		}
		// bp should equal len+1
		if ( bp != len+1 ) {
			System.err.println("parse bp="+bp+", len="+len);
		}
		os.write( buf );
	}
	/**
		@param portal null for unnamed portal, a name otherwise
		@param statementName null for unnamed portal, a name previously parse()d otherwise 
		@param paramFormats 0 = text, 1 = binary, {0} = all text, {1} = all binary, null = all text
		@param parameters data in appropriate format
		@param resultFormats 0 = text, 1 = binary, {0} = all text, {1} = all binary, null = all text */
	public void bind( String portal, String statementName,
					  short[] paramFormats, byte[][] parameters,
					  short[] resultFormats ) throws Exception {
		bindOrFunc( portal, statementName, paramFormats, parameters, resultFormats, (byte)'B' );
	}
	/**
		@param portal null for unnamed portal, a name otherwise
		@param statementName null for unnamed portal, a name previously parse()d otherwise 
		@param paramFormats 0 = text, 1 = binary, {0} = all text, {1} = all binary, null = all text
		@param parameters data in appropriate format
		@param resultFormats 0 = text, 1 = binary, {0} = all text, {1} = all binary, null = all text */
	public void functionCall( String portal, String statementName,
					  short[] paramFormats, byte[][] parameters,
					  short[] resultFormats ) throws Exception {
		bindOrFunc( portal, statementName, paramFormats, parameters, resultFormats, (byte)'F' );
	}
	/**
		@param portal null for unnamed portal, a name otherwise
		@param statementName null for unnamed portal, a name previously parse()d otherwise 
		@param paramFormats 0 = text, 1 = binary, {0} = all text, {1} = all binary, null = all text
		@param parameters data in appropriate format
		@param resultFormats 0 = text, 1 = binary, {0} = all text, {1} = all binary, null = all text
		@param op 'B' or 'F' */
	public void bindOrFunc( String portal, String statementName,
					short[] paramFormats, byte[][] parameters,
					short[] resultFormats, byte op ) throws Exception {
		int len = 4;
		byte[] p = null;
		byte[] sn = null;
		if ( portal != null ) {
			p = portal.getBytes("UTF8");
			len += p.length;
		}
		len++;
		if ( statementName != null ) {
			sn = statementName.getBytes("UTF8");
			len += sn.length;
		}
		len++;
		len += 2;
		if ( paramFormats != null ) {
			len += 2 * paramFormats.length;
		}
		len += 2;
		if ( parameters != null ) {
			for ( int i = 0; i < parameters.length; i++ ) {
				len += 4;
				if ( parameters[i] != null ) {
					len += parameters[i].length;
				}
			}
		}
		len += 2;
		if ( resultFormats != null ) {
			len += 2 * resultFormats.length;
		}
		byte[] buf = new byte[len+1];
		int bp = 0;
		buf[bp] = op;
		bp++;
		intToByteArray( buf, bp, len );
		bp += 4;
		if ( p != null ) {
			System.arraycopy( p, 0, buf, bp, p.length );
			bp += p.length;
		}
		buf[bp] = 0; bp++;
		if ( sn != null ) {
			System.arraycopy( sn, 0, buf, bp, sn.length );
			bp += sn.length;
		}
		buf[bp] = 0; bp++;
		if ( paramFormats != null ) {
			shortToByteArray( buf, bp, paramFormats.length );
			bp += 2;
			for ( int i = 0; i < paramFormats.length; i++ ) {
				shortToByteArray( buf, bp, paramFormats[i] );
				bp += 2;
			}
		} else {
			shortToByteArray( buf, bp, 0 );
			bp += 2;
		}
		if ( parameters != null ) {
			shortToByteArray( buf, bp, parameters.length );
			bp += 2;
			for ( int i = 0; i < parameters.length; i++ ) {
				if ( parameters[i] == null ) {
					intToByteArray( buf, bp, -1 );
					bp += 4;
				} else {
					intToByteArray( buf, bp, parameters[i].length );
					bp += 4;
					System.arraycopy( parameters[i], 0, buf, bp, parameters[i].length );
					bp += parameters[i].length;
				}
			}
		} else {
			shortToByteArray( buf, bp, 0 );
			bp += 2;
		}
		if ( resultFormats != null ) {
			shortToByteArray( buf, bp, resultFormats.length );
			bp += 2;
			for ( int i = 0; i < resultFormats.length; i++ ) {
				shortToByteArray( buf, bp, resultFormats[i] );
				bp += 2;
			}
		} else {
			shortToByteArray( buf, bp, 0 );
			bp += 2;
		}
		// bp should equal len+1
		System.err.println("bind: "+hexdump.hexString(buf));
		os.write( buf );
	}
	public static final byte describeStatement = (byte)'S';
	public static final byte describePortal = (byte)'P';
	/**
		@param mode describeStatement or describePortal
		@param name null for unnamed statement or portal, otherwise a name parse()d or bind()ed
		*/
	public void describe( byte mode, String name ) throws Exception {
		int len = 5; // len + mode
		byte[] n;
		if ( name != null ) {
			n = name.getBytes("UTF8");
			len += n.length + 1;
		} else {
			n = null;
			len++;
		}
		byte[] buf = new byte[len+1];
		int bp = 0;
		buf[bp] = (byte)'D';
		bp++;
		intToByteArray( buf, bp, len );
		bp += 4;
		buf[bp] = mode;
		bp++;
		if ( n != null ) {
			System.arraycopy( n, 0, buf, bp, n.length );
			bp += n.length;
		}
		buf[bp] = 0;
		os.write( buf );
	}
	/**
		execute a bound portal
		@param name null for unnamed portal, otherwise a name previously bind()ed
		@param rowlimit 0 for no limit, otherwise the maximum number of rows to return
		*/
	public void execute( String name, int rowlimit ) throws Exception {
		int len = 8; // len + rowlimit
		byte[] n;
		if ( name != null ) {
			n = name.getBytes("UTF8");
			len += n.length + 1;
		} else {
			n = null;
			len++;
		}
		byte[] buf = new byte[len+1];
		int bp = 0;
		buf[bp] = (byte)'E';
		bp++;
		intToByteArray( buf, bp, len );
		bp += 4;
		if ( n != null ) {
			System.arraycopy( n, 0, buf, bp, n.length );
			bp += n.length;
		}
		buf[bp] = 0;
		bp++;
		intToByteArray( buf, bp, rowlimit );
		bp += 4;
		// bp should equal len+1
		os.write( buf );
	}
	public void close( byte mode, String name ) throws Exception {
		int len = 5; // len + mode
		byte[] n;
		if ( name != null ) {
			n = name.getBytes("UTF8");
			len += n.length + 1;
		} else {
			n = null;
			len++;
		}
		byte[] buf = new byte[len+1];
		int bp = 0;
		buf[bp] = (byte)'D';
		bp++;
		intToByteArray( buf, bp, len );
		bp += 4;
		buf[bp] = mode;
		bp++;
		if ( n != null ) {
			System.arraycopy( n, 0, buf, bp, n.length );
			bp += n.length;
		}
		buf[bp] = 0;
		os.write( buf );
	}
	/** send sync message */
	public void sync() throws Exception {
		os.write( syncMessage );
	}
	/** send CancelRequest messsage
		@param pid data received at connect time
		@param key data received at connect time */
	public void cancelRequest( int pid, int key ) throws Exception {
		byte[] buf = new byte[16];
		int bp = 0;
		intToByteArray( buf, bp, 16 );
		bp += 4;
		intToByteArray( buf, bp, 80877102 ); // 0x12345678
		bp += 4;
		intToByteArray( buf, bp, pid );
		bp += 4;
		intToByteArray( buf, bp, key );
		os.write( buf );
	}
	/** send SSLRequest message */
	public void sslRequest() throws Exception {
		byte[] buf = new byte[8];
		int bp = 0;
		intToByteArray( buf, bp, 8 );
		bp += 4;
		intToByteArray( buf, bp, 80877103 ); // 0x12345679
		os.write( buf );
	}
	/** sends data. no copy is performed.
		@param data sent in a CopyData message to db server
		*/
	public void copyData( byte[] data ) throws Exception {
		byte[] head = new byte[5];
		head[0] = (byte)'d';
		intToByteArray( head, 1, data.length + 4 );
		os.write( head );
		os.write( data );
	}
	/** sends a copy data header. Write the rest on the OutputStream returned.
		@param len length of data to copy
		@return a stream to write the data to. do not close() the stream. do flush() the stream.
		*/
	public OutputStream copyDataStart( int len ) throws Exception {
		byte[] head = new byte[5];
		head[0] = (byte)'d';
		intToByteArray( head, 1, len + 4 );
		os.write( head );
		return os;
	}
	/** sends CopyDone message */
	public void copyDone() throws Exception {
		byte[] head = new byte[5];
		head[0] = (byte)'c';
		intToByteArray( head, 1, 4 );
		os.write( head );
	}
	/** sends CopyFail message
		@param message description of error
		*/
	public void copyFail( String message ) throws Exception {
		sendString( message, (byte)'f' );
	}
	/** sends PasswordMessage message
		@param message password
		*/
	public void passwordMessage( String message ) throws Exception {
		sendString( message, (byte)'p' );
	}
	/** simple Query
		@param message sql text
		*/
	public void query( String message ) throws Exception {
		sendString( message, (byte)'Q' );
	}
	/** send a string for various simple operations
		@param message string to send
		@param op f,p,Q,(future protocol expansion?)
		*/
	public void sendString( String message, byte op ) throws Exception {
		byte[] msg = message.getBytes("UTF8");
		byte[] buf = new byte[6+msg.length];
		buf[0] = op;
		intToByteArray( buf, 1, msg.length + 4 );
		System.arraycopy( msg, 0, buf, 5, msg.length );
		buf[buf.length-1] = 0;
		os.write( buf );
	}
	/** sends Flush message */
	public void flush() throws Exception {
		byte[] head = new byte[5];
		head[0] = (byte)'H';
		intToByteArray( head, 1, 4 );
		os.write( head );
	}
	
	LinkedList returnedChunks = new LinkedList();
	
	/** how many ProtocolChunks do we have from the server?
		@return number of stored chunks
		*/
	public int availableChunks() {
		synchronized ( returnedChunks ) {
			return returnedChunks.size();
		}
	}
	/** add a chunk to the queue.
		probably just called by the thread reading from the socked, but could be used to inject other traffic.
		@param it a ProtocolChunk
		*/
	public void addChunk( ProtocolChunk it ) {
		synchronized ( returnedChunks ) {
			returnedChunks.add( it );
			returnedChunks.notify();
		}
	}
	/**
		get the first chunk off the queue.
	 Blocks for up to timeoutMillis if there is none
	 @param timeoutMillis max millis to wait()
	 @return first available ProtocolChunk from the queue*/
	public ProtocolChunk getChunk( long timeoutMillis ) {
		synchronized ( returnedChunks ) {
			if ( returnedChunks.size() == 0 ) {
				try {
					returnedChunks.wait( timeoutMillis, 0 );
				} catch ( InterruptedException e ) {
					return null;
				}
			}
			return (ProtocolChunk)returnedChunks.getFirst();
		}
	}

	boolean hexdumpServerStream = false;
	boolean outPrintChunks = true;
	public void run() {
		byte[] buf = new byte[1024];
		int len;
		PrintWriter out;
		PipedOutputStream pos = null;
		DataInputStream qdis;
		try {
			out = new PrintWriter( System.out );
			if ( hexdumpServerStream ) {
				pos = new PipedOutputStream();
				qdis = new DataInputStream( new PipedInputStream( pos ));
			} else {
				qdis = is;
			}
			(new ChunkReadThread( qdis, returnedChunks, outPrintChunks?out:null)).start();
		} catch ( Exception e ) {
			e.printStackTrace();
			return;
		}
		while ( hexdumpServerStream ) {
			try {
				len = is.read( buf );
				if ( len > 0 ) {
					hexdump.print( out, buf, len, 0 );
					out.flush();
					pos.write( buf, 0, len );
				} else if ( len == -1 ) {
					// eof
					break;
				} else {
					System.out.println("closed?");
				}
			} catch ( Exception e ) {
				e.printStackTrace();
				break;
			}
		}
		if ( pos != null ) try {
			pos.close();
		} catch ( Exception e ) {
			e.printStackTrace();
		}
		System.out.println("read thread exits");
	}

    public static void main (String args[]) {
		pgj it = new pgj();
        System.out.println("begin pgj toy test");
		long ms = 5000;
		try {
			it.openConnection( defaultServerName, defaultServerPort, new String[]{ "user","poll","client_encoding","UNICODE" });
			System.out.println("connection opened");
			Thread.sleep(ms);
			it.parse( null, "SELECT * FROM polls", null );
			System.out.println("parse sent");
			Thread.sleep(ms);
			it.bind( null, null, null, null, new short[]{ 1 } );
			System.out.println("bind");
			Thread.sleep(ms);
			it.describe( describePortal, null );
			System.out.println("describe portal");
			Thread.sleep(ms);
			it.execute( null, 0 );
			System.out.println("execute");
			Thread.sleep(ms);
			it.close( describePortal, null );
			System.out.println("close");
			Thread.sleep(ms);
			it.sync();
			System.out.println("sync");
			Thread.sleep(ms);
			it.closeConnection();
		} catch ( Exception e ) {
			e.printStackTrace();
		}
        System.out.println("done");
    }
	
	/** tagging interface.
		Is there any useful operation that is generic across all the protocol chunks?
		Data is pretty specific. Type shall be determined by 'instanceof'. */
	public interface ProtocolChunk {
	}
	/** internal struct for organizing the parser */
	static class ProtocolReaderListEntry {
		byte key;
		ProtocolReader r;
		public ProtocolReaderListEntry( byte k, ProtocolReader rin ) {
			key = k;
			r = rin;
		}
	}
	/** field of internal struct for organizing the parser */
	interface ProtocolReader {
		/** the initial byte has been read, read the body */
		ProtocolChunk read( DataInputStream dis ) throws java.io.IOException;
	}
	public static class AuthReply implements ProtocolChunk {
		public int reply;
		public byte salt[];
		public static final int OK = 0;
		public static final int KERB4 = 1;
		public static final int KERB5 = 2;
		public static final int TEXT = 3;
		public static final int CRYPT = 4;
		public static final int MD5 = 5;
		public static final int SCM = 6;
		public static final String[] replyDescriptions = new String[] {
			"Ok",
			"KerberosV4 Reqired",
			"KerberosV5 Reqired",
			"ClearText Password Reqired", 
			"Crypt Password Reqired", 
			"MD5 Password Reqired", 
			"SCM Credential Reqired", 
		};
		AuthReply( int r, byte[] s ) {
			reply = r;
		}
		public String toString() {
			if ( salt != null ) {
				return "AuthReply: " + replyDescriptions[reply] + "&salt=" + hexdump.hexString( salt );
			} else {
				return "AuthReply: " + replyDescriptions[reply];
			}
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				int reply = dis.readInt();
				byte[] s = null;
				switch ( reply ) {
					case 4:
						s = new byte[2];
						dis.readFully( s );
						// size should equal 10
						break;
					case 5:
						s = new byte[4];
						dis.readFully( s );
						// size should equal 12
						break;
					default:
						// size should equal 8
						break;
				}
				return new AuthReply( reply, s );
			}
		}
	}
	public static class BackendKeyData implements ProtocolChunk {
		public int pid;
		public int key;
		public BackendKeyData( int p, int k ) {
			pid = p;
			key = k;
		}
		public String toString() {
			return "BackendKeyData: " + pid + "=" + key;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				int pid = dis.readInt();
				int key = dis.readInt();
				// size should equal 12
				return new BackendKeyData( pid, key );
			}
		}
	}
	public static class BindComplete implements ProtocolChunk {
		public BindComplete() {
		}
		public String toString() {
			return "BindComplete";
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new BindComplete();
			}
		}
	}
	public static class CloseComplete implements ProtocolChunk {
		public CloseComplete() {
		}
		public String toString() {
			return "CloseComplete";
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new CloseComplete();
			}
		}
	}
	public static class CommandComplete implements ProtocolChunk {
		public String tag;
		public CommandComplete( String t ) {
			tag = t;
		}
		public String toString() {
			return "CommandComplete: "+tag;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				byte[] buf = new byte[size - 4];
				dis.readFully( buf );
				return new CommandComplete( new String( buf ) );
			}
		}
	}
	public static class CopyData implements ProtocolChunk {
		public byte[] data;
		public CopyData( byte[] d ) {
			data = d;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				byte[] buf = new byte[size - 4];
				dis.readFully( buf );
				return new CopyData( buf );
			}
		}
	}
	public static class CopyDone implements ProtocolChunk {
		public CopyDone() {
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new CopyDone();
			}
		}
	}
	// copy in/out response
	public static class DataRow implements ProtocolChunk {
		public byte[][] data;
		public DataRow( byte[][] d ) {
			data = d;
		}
		public String toString() {
			StringBuffer toret = new StringBuffer( "DataRow: " );
			for ( int i = 0; i < data.length; i++ ) {
				if ( data[i] != null ) {
					toret.append(hexdump.hexString(data[i]));
				}
				toret.append(", ");
			}
			return toret.toString();
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				int sizecheck = 4;
				short cols = dis.readShort();
				sizecheck += 2;
				byte[][] data = new byte[cols][];
				for ( short i = 0; i < cols; i++ ) {
					int colsize;
					byte[] buf;
					colsize = dis.readInt();
					sizecheck += 4;
					if ( colsize == -1 ) {
						data[i] = null;
					} else if ( colsize == 0 ) {
						data[i] = new byte[0];
					} else {
						data[i] = new byte[colsize];
						dis.readFully( data[i] );
						sizecheck += colsize;
					}
				}
				// sizecheck shoul dequal size
				return new DataRow( data );
			}
		}
	}
	public static class EmitQueryResponse implements ProtocolChunk {
		public EmitQueryResponse() {
		}
		public String toString() {
			return "EmitQueryResponse";
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new EmitQueryResponse();
			}
		}
	}
	public static class ErrorResponse implements ProtocolChunk {
		public String[] fields;
		public ErrorResponse( String[] f ) {
			fields = f;
		}
		public String toString() {
			StringBuffer toret = new StringBuffer("ErrorResponse: ");
			if ( fields != null ) for ( int i = 0; i < fields.length; i++ ) {
				toret.append( fields[i] ).append( ", " );
			}
			return  toret.toString();
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				byte[] buf = new byte[size - 4];
				dis.readFully( buf );
				int chunks = 0;
				int bp;
				for ( bp = 0; bp < buf.length; bp++ ) {
					if ( buf[bp] == 0 ) {
						chunks++;
					}
				}
				if ( buf[buf.length - 1] != 0 ) {
					chunks++;
				}
				String[] f = new String[chunks];
				int cs = 0;
				int fi = 0;
				for ( bp = 0; bp < buf.length; bp++ ) {
					if ( buf[bp] == 0 ) {
						f[fi] = new String( buf, cs, bp - cs );
						cs = bp + 1;
						fi++;
					}
				}
				if ( cs < buf.length ) {
					f[fi] = new String( buf, cs, buf.length - cs );
				}
				// rl should equal size
				return new ErrorResponse( f );
			}
		}
		public static byte Severity = (byte)'S';
		public static byte Code = (byte)'C';
		public static byte Message = (byte)'M';
		public static byte Detail = (byte)'D';
		public static byte Hint = (byte)'H';
		public static byte Position = (byte)'P';
		public static byte Where = (byte)'W';
		public static byte File = (byte)'F';
		public static byte Line = (byte)'L';
		public static byte Routine = (byte)'R';
	}
	public static class FunctionCallResponse implements ProtocolChunk {
		public byte[] data;
		public FunctionCallResponse( byte[] d ) {
			data = d;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				int len = dis.readInt();
				if ( len == -1 ) {
					return new FunctionCallResponse( null );
				} else if ( len == 0 ) {
					return new FunctionCallResponse( new byte[0] );
				} else {
					byte[] buf = new byte[len];
					dis.readFully( buf );
					return new FunctionCallResponse( buf );
				}
			}
		}
	}
	public static class NoData implements ProtocolChunk {
		public NoData() {
		}
		public String toString() {
			return "NoData";
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new NoData();
			}
		}
	}
	public static class NotificationResponse implements ProtocolChunk {
		public int pid;
		public String name;
		public String info;
		public NotificationResponse( String t, String i ) {
			name = t;
			info = i;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				int pid = dis.readInt();
				byte[] buf = new byte[size - 8];
				dis.readFully( buf );
				int split = 0;
				while ( buf[split] != 0 ) {
					split++;
				}
				return new NotificationResponse( new String( buf, 0, split-1 ), new String( buf, split+1, buf.length - (split+1) ) );
			}
		}
	}
	public static class ParameterDescription implements ProtocolChunk {
		public int types[];
		public ParameterDescription( int[] t ) {
			types = t;
		}
		public String toString() {
			StringBuffer toret = new StringBuffer( "ParameterDescription: " );
			for ( int i = 0; i < types.length; i++ ) {
				toret.append( i );
				if ( i + 1 < types.length ) {
					toret.append(", ");
				}
			}
			return toret.toString();
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				short nump = dis.readShort();
				int t[] = new int[nump];
				for ( int i = 0; i < nump; i++ ) {
					t[i] = dis.readInt();
				}
				// size should equal 6 + nump*4
				return new ParameterDescription( t );
			}
		}
	}
	public static class ParameterStatus implements ProtocolChunk {
		public int pid;
		public String name;
		public String value;
		public ParameterStatus( String t, String i ) {
			name = t;
			value = i;
		}
		public String toString() {
			return "ParameterStatus: " + name + "=" + value;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				byte[] buf = new byte[size - 4];
				dis.readFully( buf );
				int split = 0;
				while ( buf[split] != 0 ) {
					split++;
				}
				return new ParameterStatus( new String( buf, 0, split ), new String( buf, split+1, buf.length - (split+1) ) );
			}
		}
	}
	public static class ParseComplete implements ProtocolChunk {
		public ParseComplete() {
		}
		public String toString() {
			return "ParseComplete";
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new ParseComplete();
			}
		}
	}
	public static class PortalSuspended implements ProtocolChunk {
		public PortalSuspended() {
		}
		public String toString() {
			return "PortalSuspended";
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				// size should equal 4
				return new PortalSuspended();
			}
		}
	}
	public static class ReadyForQuery implements ProtocolChunk {
		public byte code;
		public ReadyForQuery( byte c ) {
			code = c;
		}
		public String toString() {
			return "ReadyForQuery: " + (char)code;
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				byte c = dis.readByte();
				// size should equal 5
				return new ReadyForQuery( c );
			}
		}
		public static final byte Idle = (byte)'I';
		public static final byte Transaction = (byte)'T';
		public static final byte Error = (byte)'E';
	}
	public static void intToByteArray( byte[] a, int pos, int i ) {
		a[pos] = (byte)((i >> 24) & 0xff);pos++;
		a[pos] = (byte)((i >> 16) & 0xff);pos++;
		a[pos] = (byte)((i >>  8) & 0xff);pos++;
		a[pos] = (byte)((i      ) & 0xff);
	}
	public static void shortToByteArray( byte[] a, int pos, int i ) {
		a[pos] = (byte)((i >>  8) & 0xff);pos++;
		a[pos] = (byte)((i      ) & 0xff);
	}
	public static int intFromByteArray( byte[] a, int p ) {
		int ba = a[p];
		ba = ba << 24;
		p++;
		int bb = a[p];
		bb = bb & 0xff;
		ba = ba | (bb << 16);
		p++;
		bb = a[p];
		bb = bb & 0xff;
		ba = ba | (bb << 8);
		p++;
		bb = a[p];
		bb = bb & 0xff;
		return ba | bb;
	}
	public static short shortFromByteArray( byte[] a, int p ) {
		int ba = a[p];
		ba = ba << 8;
		p++;
		int bb = a[p];
		bb = bb & 0xff;
		return (short)(ba | bb);
	}
	public static class RowDescription implements ProtocolChunk {
		public static class Field {
			String name;
			int tableOID;
			short attr;
			int type;
			short typeLen;
			int typeModifier;
			/** 0 = text, 1 = binary */
			short format;
		}
		public Field[] fields;
		public RowDescription( Field[] f ) {
			fields = f;
		}
		public String toString() {
			StringBuffer toret = new StringBuffer("RowDescription: ");
			if ( fields != null ) for ( int i = 0; i < fields.length; i++ ) {
				toret.append("(name=").append(fields[i].name).append(" table=").append(Integer.toHexString(fields[i].tableOID)).append(" attr=").append(Integer.toHexString(fields[i].attr)).append(" type=").append(fields[i].type).append(" typeLen=").append(fields[i].typeLen).append(" typemod=").append(Integer.toHexString(fields[i].typeModifier));
				if ( fields[i].format == 0 ) {
					toret.append(" text)");
				} else if ( fields[i].format == 1 ) {
					toret.append(" binary)");
				} else {
					toret.append(" ").append(fields[i].format).append(")");
				}
				if ( i + 1 < fields.length ) {
					toret.append(", ");
				}
			}
			return toret.toString();
		}
		static class Reader implements ProtocolReader {
			public ProtocolChunk read( DataInputStream dis ) throws java.io.IOException {
				int size = dis.readInt();
				short numf = dis.readShort();
				byte[] bin = new byte[size-6];
				int bp = 0;
				dis.readFully( bin );
				Field[] f = new Field[numf];
				for ( int i = 0; i < numf; i++ ) {
					f[i] = new Field();
					{
						int fns = bp;
						while ( bin[bp] != 0 ) {
							bp++;
						}
						f[i].name = new String( bin, fns, bp - fns );
						bp++;
					}
					f[i].tableOID = intFromByteArray( bin, bp );
					bp += 4;
					f[i].attr = shortFromByteArray( bin, bp );
					bp += 2;
					f[i].type = intFromByteArray( bin, bp );
					bp += 4;
					f[i].typeLen = shortFromByteArray( bin, bp );
					bp += 2;
					f[i].typeModifier = intFromByteArray( bin, bp );
					bp += 4;
					f[i].format = shortFromByteArray( bin, bp );
					bp += 2;
				}
				// bp should equal size-6
				return new RowDescription(f);
			}
		}
	}
	static final ProtocolReaderListEntry prot[] = new ProtocolReaderListEntry[] {
		new ProtocolReaderListEntry( (byte)'R', new AuthReply.Reader() ),
		new ProtocolReaderListEntry( (byte)'K', new BackendKeyData.Reader() ),
		new ProtocolReaderListEntry( (byte)'2', new BindComplete.Reader() ),
		new ProtocolReaderListEntry( (byte)'3', new CloseComplete.Reader() ),
		new ProtocolReaderListEntry( (byte)'C', new CommandComplete.Reader() ),
		new ProtocolReaderListEntry( (byte)'d', new CopyData.Reader() ),
		new ProtocolReaderListEntry( (byte)'c', new CopyDone.Reader() ),
		new ProtocolReaderListEntry( (byte)'D', new DataRow.Reader() ),
		new ProtocolReaderListEntry( (byte)'l', new EmitQueryResponse.Reader() ),
		new ProtocolReaderListEntry( (byte)'E', new ErrorResponse.Reader() ),
		new ProtocolReaderListEntry( (byte)'V', new FunctionCallResponse.Reader() ),
		new ProtocolReaderListEntry( (byte)'n', new NoData.Reader() ),
		new ProtocolReaderListEntry( (byte)'N', new ErrorResponse.Reader() ),
		new ProtocolReaderListEntry( (byte)'A', new NotificationResponse.Reader() ),
		new ProtocolReaderListEntry( (byte)'t', new ParameterDescription.Reader() ),
		new ProtocolReaderListEntry( (byte)'S', new ParameterStatus.Reader() ),
		new ProtocolReaderListEntry( (byte)'1', new ParseComplete.Reader() ),
		new ProtocolReaderListEntry( (byte)'s', new PortalSuspended.Reader() ),
		new ProtocolReaderListEntry( (byte)'Z', new ReadyForQuery.Reader() ),
		new ProtocolReaderListEntry( (byte)'T', new RowDescription.Reader() ),
//		new ProtocolReaderListEntry( (byte)'', new .Reader() ),
	};
	static ProtocolChunk readChunk( DataInputStream is ) throws java.io.IOException {
		byte key = is.readByte();
		// fixme, sort prot on key and do binary search, OR use a 128 element LUT
		for ( int i = 0; i < prot.length; i++ ) {
			if ( prot[i].key == key ) {
				return prot[i].r.read( is );
			}
		}
		return null;
	}
	/** Reads and parses input stream, storing to queue and printing. */
	public static class ChunkReadThread extends Thread {
		DataInputStream is;
		PrintWriter out = null;
		LinkedList q = null;
		/**
		 @param d DataInputStream likely from a Socket connected to a database.
		 @param chunkQueue where to store parsed chunks to. May be null.
		 @param debugOut println() to this for each chunk's toString(). May be null.
		 */
		ChunkReadThread( DataInputStream d, LinkedList chunkQueue, PrintWriter debugOut ) {
			is = d;
			q = chunkQueue;
			out = debugOut;
		}
		/**
		 @param d DataInputStream likely from a Socket connected to a database.
		 @param chunkQueue where to store parsed chunks to.
		 */
		ChunkReadThread( DataInputStream d, LinkedList chunkQueue ) {
			is = d;
			q = chunkQueue;
			out = null;
		}
		/**
		 @param d DataInputStream likely from a Socket connected to a database.
		 @param debugOut println() to this for each chunk's toString().
		 */
		ChunkReadThread( DataInputStream d, PrintWriter debugOut ) {
			is = d;
			q = null;
			out = debugOut;
		}
		/** Reads and parses input stream, storing to queue and printing. */
		public void run() {
			try {
				while ( true ) {
					ProtocolChunk pc;
					pc = pgj.readChunk( is );
					if ( pc != null ) {
						if ( q != null ) synchronized ( q ) {
							q.add( pc );
							q.notify();
						}
						if ( out != null ) {
							out.println( pc );
							out.flush();
						}
					}
				}
			} catch ( Exception e ) {
				if ( out != null ) {
					e.printStackTrace( out );
				} else {
					//e.printStackTrace();
				}
			}
		}
	}
}
